#! /usr/bin/env python3
# Copyright (c) 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getopt
import sys


def strip_dpdk(check_dpdk, src, dst):
    disabled_print = False
    while True:
        line = src.readline()
        if not line:
            break
        if '@begin_dpdk@' in line or '@end_dpdk@' in line:
            if not check_dpdk:
                disabled_print = not disabled_print
            continue
        if not disabled_print:
            dst.write(line)


if __name__ == '__main__':
    check_dpdk = False
    options, args = getopt.gnu_getopt(sys.argv[1:], '', ['dpdk', 'nodpdk'])
    for key, value in options:
        if key == '--dpdk':
            check_dpdk = True
        elif key == '--nodpdk':
            check_dpdk = False
        else:
            assert False
    if args:
        for arg in args:
            strip_dpdk(check_dpdk, open(arg), sys.stdout)
    else:
        strip_dpdk(check_dpdk, sys.stdin, sys.stdout)



# Copyright (c) 2011 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import ovs.vlog


def main():
    modules = [ovs.vlog.Vlog("module_%d" % i) for i in range(3)]

    parser = argparse.ArgumentParser(description="Vlog Module Tester")
    ovs.vlog.add_args(parser)
    args = parser.parse_args()
    ovs.vlog.handle_args(args)

    for m in modules:
        m.emer("emergency")
        m.err("error")
        m.warn("warning")
        m.info("information")
        m.dbg("debug")

        try:
            fail = False  # Silence pychecker warning.
            assert fail
        except AssertionError:
            m.emer("emergency exception", exc_info=True)
            m.err("error exception", exc_info=True)
            m.warn("warn exception", exc_info=True)
            m.info("information exception", exc_info=True)
            m.dbg("debug exception", exc_info=True)
            m.exception("exception")


if __name__ == '__main__':
    main()



# Copyright (c) 2009, 2010, 2012 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import sys

import ovs.reconnect

now = 0
r = None


def do_enable(_):
    r.enable(now)


def do_disable(_):
    r.disable(now)


def do_force_reconnect(_):
    r.force_reconnect(now)


def error_from_string(s):
    if not s:
        return 0
    elif s == "ECONNREFUSED":
        return errno.ECONNREFUSED
    elif s == "EOF":
        return ovs.reconnect.EOF
    else:
        sys.stderr.write("unknown error '%s'\n" % s)
        sys.exit(1)


def do_disconnected(arg):
    r.disconnected(now, error_from_string(arg))


def do_connecting(_):
    r.connecting(now)


def do_connect_failed(arg):
    r.connect_failed(now, error_from_string(arg))


def do_connected(_):
    r.connected(now)


def do_activity(_):
    r.activity(now)


def do_run(arg):
    global now
    if arg is not None:
        now += int(arg)

    action = r.run(now)
    if action is None:
        pass
    elif action == ovs.reconnect.CONNECT:
        print("  should connect")
    elif action == ovs.reconnect.DISCONNECT:
        print("  should disconnect")
    elif action == ovs.reconnect.PROBE:
        print("  should send probe")
    else:
        assert False


def do_advance(arg):
    global now
    now += int(arg)


def do_timeout(_):
    global now
    timeout = r.timeout(now)
    if timeout is not None and timeout >= 0:
        print("  advance %d ms" % timeout)
        now += timeout
    else:
        print("  no timeout")


def do_set_max_tries(arg):
    r.set_max_tries(int(arg))


def do_set_backoff_free_tries(arg):
    r.set_backoff_free_tries(int(arg))


def diff_stats(old, new, delta):
    if (old.state != new.state or
        old.state_elapsed != new.state_elapsed or
        old.backoff != new.backoff):
        print("  in %s for %d ms (%d ms backoff)"
              % (new.state, new.state_elapsed, new.backoff))

    if (old.creation_time != new.creation_time or
        old.last_activity != new.last_activity or
        old.last_connected != new.last_connected):
        print("  created %d, last activity %d, last connected %d"
              % (new.creation_time, new.last_activity, new.last_connected))

    if (old.n_successful_connections != new.n_successful_connections or
        old.n_attempted_connections != new.n_attempted_connections or
        old.seqno != new.seqno):
        print("  %d successful connections out of %d attempts, seqno %d"
              % (new.n_successful_connections, new.n_attempted_connections,
                 new.seqno))

    if (old.is_connected != new.is_connected):
        if new.is_connected:
            negate = ""
        else:
            negate = "dis"
        print("  %sconnected" % negate)

    if (old.last_connected != new.last_connected or
        (new.msec_since_connect is not None and
         old.msec_since_connect != new.msec_since_connect - delta) or
        (old.total_connected_duration != new.total_connected_duration - delta
            and not (old.total_connected_duration == 0 and
                new.total_connected_duration == 0))):
        print("  last connected %d ms ago, connected %d ms total"
              % (new.msec_since_connect, new.total_connected_duration))

    if (old.last_disconnected != new.last_disconnected or
        (new.msec_since_disconnect is not None and
         old.msec_since_disconnect != new.msec_since_disconnect - delta)):
        print("  disconnected at %d ms (%d ms ago)"
              % (new.last_disconnected, new.msec_since_disconnect))


def do_set_passive(_):
    r.set_passive(True, now)


def do_listening(_):
    r.listening(now)


def do_listen_error(arg):
    r.listen_error(now, int(arg))


def do_receive_attempted(arg):
    if arg == "now":
        r.receive_attempted(now)
    elif arg == "LLONG_MAX":
        r.receive_attempted(None)
    else:
        sys.stderr.write("receive-attempted: bad argument %s\n" % arg)
        sys.exit(1)


def main():
    commands = {
        "enable": do_enable,
        "disable": do_disable,
        "force-reconnect": do_force_reconnect,
        "disconnected": do_disconnected,
        "connecting": do_connecting,
        "connect-failed": do_connect_failed,
        "connected": do_connected,
        "activity": do_activity,
        "run": do_run,
        "advance": do_advance,
        "timeout": do_timeout,
        "set-max-tries": do_set_max_tries,
        "set-backoff-free-tries": do_set_backoff_free_tries,
        "passive": do_set_passive,
        "listening": do_listening,
        "listen-error": do_listen_error,
        "receive-attempted": do_receive_attempted
    }

    global now
    global r

    now = 1000
    r = ovs.reconnect.Reconnect(now)
    r.set_name("remote")
    prev = r.get_stats(now)
    print("### t=%d ###" % now)
    old_time = now
    old_max_tries = r.get_max_tries()
    while True:
        line = sys.stdin.readline()
        if line == "":
            break

        print(line[:-1])
        if line[0] == "#":
            continue

        args = line.split()
        if len(args) == 0:
            continue

        command = args[0]
        if len(args) > 1:
            op = args[1]
        else:
            op = None
        commands[command](op)

        if old_time != now:
            print()
            print("### t=%d ###" % now)

        cur = r.get_stats(now)
        diff_stats(prev, cur, now - old_time)
        prev = cur
        if r.get_max_tries() != old_max_tries:
            old_max_tries = r.get_max_tries()
            print("  %d tries left" % old_max_tries)

        old_time = now


if __name__ == '__main__':
    main()



# Copyright (c) 2010, 2011, 2012, 2015, 2016, 2017 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys

from ovs.db import error


def text_to_nroff(s, font=r'\fR', escape_dot=True):
    def escape(match):
        c = match.group(0)

        # In Roman type, let -- in XML be \- in nroff.  That gives us a way to
        # write minus signs, which is important in some places in manpages.
        #
        # Bold in nroff usually represents literal text, where there's no
        # distinction between hyphens and minus sign.  The convention in nroff
        # appears to be to use a minus sign in such cases, so we follow that
        # convention.
        #
        # Finally, we always output - as a minus sign when it is followed by a
        # digit.
        if c.startswith('-'):
            if c == '--' and font == r'\fR':
                return r'\-'
            if c != '-' or font in (r'\fB', r'\fL'):
                return c.replace('-', r'\-')
            else:
                return '-'

        if c == '\\':
            return r'\e'
        elif c == '"':
            return r'\(dq'
        elif c == "'":
            return r'\(cq'
        elif c == ".":
            if escape_dot:
                # groff(7) says that . can be escaped by \. but in practice
                # groff still gives an error with \. at the beginning of a
                # line.
                return r'\[char46]'
            else:
                return '.'
        else:
            raise error.Error("bad escape")

    # Escape - \ " ' . as needed by nroff.
    s = re.sub('(-[0-9]|--|[-"\'\\\\.])', escape, s)
    return s


def escape_nroff_literal(s, font=r'\fB'):
    return font + r'%s\fR' % text_to_nroff(s, font)


def inline_xml_to_nroff(node, font, to_upper=False, newline='\n'):
    if node.nodeType == node.TEXT_NODE:
        if to_upper:
            s = text_to_nroff(node.data.upper(), font)
        else:
            s = text_to_nroff(node.data, font)
        return s.replace('\n', newline)
    elif node.nodeType == node.ELEMENT_NODE:
        if node.tagName in ['code', 'em', 'option', 'env', 'b']:
            s = r'\fB'
            for child in node.childNodes:
                s += inline_xml_to_nroff(child, r'\fB', to_upper, newline)
            return s + font
        elif node.tagName == 'ref':
            if node.hasAttribute('column'):
                s = node.attributes['column'].nodeValue
                if node.hasAttribute('key'):
                    s += ':' + node.attributes['key'].nodeValue
            elif node.hasAttribute('table'):
                s = node.attributes['table'].nodeValue
            elif node.hasAttribute('group'):
                s = node.attributes['group'].nodeValue
            elif node.hasAttribute('db'):
                s = node.attributes['db'].nodeValue
            elif node.hasAttribute('field'):
                s = node.attributes['field'].nodeValue
            elif node.hasAttribute('section'):
                s = node.attributes['section'].nodeValue
            else:
                raise error.Error("'ref' lacks required attributes: %s"
                                  % list(node.attributes.keys()))
            return r'\fB' + re.sub(r'\s+', ' ', s) + font
        elif node.tagName in ['var', 'dfn', 'i', 'cite']:
            s = r'\fI'
            for child in node.childNodes:
                s += inline_xml_to_nroff(child, r'\fI', to_upper, newline)
            return s + font
        elif node.tagName in ['literal']:
            s = r'\fL'
            for child in node.childNodes:
                s += inline_xml_to_nroff(child, r'\fL')
            return s + font
        elif node.tagName == 'url':
            return ('\n.URL "'
                    + text_to_nroff(node.attributes['href'].nodeValue,
                                    escape_dot=False)
                    + '"\n')
        else:
            raise error.Error("element <%s> unknown or invalid here"
                              % node.tagName)
    elif node.nodeType == node.COMMENT_NODE:
        return ''
    else:
        raise error.Error("unknown node %s in inline xml" % node)


def pre_to_nroff(nodes, para, font):
    # This puts 'font' at the beginning of each line so that leading and
    # trailing whitespace stripping later doesn't removed leading spaces
    # from preformatted text.
    s = para + '\n.nf\n' + font
    for node in nodes:
        s += inline_xml_to_nroff(node, font, False, '\n.br\n' + font) + '\\fR'
    s += '\n.fi\n'
    return s


def tbl_to_nroff(nodes, para):
    s = para + '\n.TS\n'
    for node in nodes:
        if node.nodeType != node.TEXT_NODE:
            fatal("<tbl> element may only have text children")
        s += node.data + '\n'
    s += '.TE\n'
    return s


def fatal(msg):
    sys.stderr.write('%s\n' % msg)
    sys.exit(1)


def put_text(text, x, y, s):
    x = int(x)
    y = int(y)
    extend = x + len(s) - len(text[y])
    if extend > 0:
        text[y] += ' ' * extend
    text[y] = text[y][:x] + s + text[y][x + len(s):]


def put_centered(text, x, width, y, s):
    put_text(text, x + (width - len(s)) / 2, y, s)


def diagram_header_to_nroff(header_node, text, x):
    # Parse header.
    header_fields = []
    i = 0
    for node in header_node.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName == 'bits':
            name = node.attributes['name'].nodeValue
            width = node.attributes['width'].nodeValue
            above = node.getAttribute('above')
            below = node.getAttribute('below')
            fill = node.getAttribute('fill')
            header_fields += [{"name": name,
                              "tag": "B%d" % i,
                              "width": width,
                              "above": above,
                              "below": below,
                              "fill": fill}]
            i += 1
        elif node.nodeType == node.COMMENT_NODE:
            pass
        elif node.nodeType == node.TEXT_NODE and node.data.isspace():
            pass
        else:
            fatal("unknown node %s in diagram <header> element" % node)

    # Format pic version.
    pic_s = ""
    for f in header_fields:
        name = f['name'].replace('...', '. . .')
        pic_s += "  %s: box \"%s\" width %s" % (f['tag'], name, f['width'])
        if f['fill'] == 'yes':
            pic_s += " fill"
        pic_s += '\n'
    for f in header_fields:
        pic_s += "  \"%s\" at %s.n above\n" % (f['above'], f['tag'])
        pic_s += "  \"%s\" at %s.s below\n" % (f['below'], f['tag'])
    name = header_node.getAttribute('name')
    if name == "":
        visible = " invis"
    else:
        visible = ""
    pic_s += "line <->%s \"%s\" above " % (visible, name)
    pic_s += "from %s.nw + (0,textht) " % header_fields[0]['tag']
    pic_s += "to %s.ne + (0,textht)\n" % header_fields[-1]['tag']

    # Format text version.
    header_width = 1
    for f in header_fields:
        field_width = max(len(f['above']), len(f['below']), len(f['name']))
        f['width'] = field_width
        header_width += field_width + 1
    min_header_width = 2 + len(name)
    while min_header_width > header_width:
        for f in header_fields:
            f['width'] += 1
            header_width += 1
            if header_width >= min_header_width:
                break

    if name != "":
        put_centered(text, x, header_width, 0, name)
        if header_width >= 4:
            arrow = '<' + '-' * (header_width - 4) + '>'
            put_text(text, x + 1, 1, arrow)
    for f in header_fields:
        box1 = '+' + '-' * f['width'] + '+'
        box2 = '|' + ' ' * f['width'] + '|'
        put_text(text, x, 3, box1)
        put_text(text, x, 4, box2)
        put_text(text, x, 5, box1)

        put_centered(text, x + 1, f['width'], 2, f['above'])
        put_centered(text, x + 1, f['width'], 4, f['name'])
        put_centered(text, x + 1, f['width'], 6, f['below'])

        x += f['width'] + 1

    return pic_s, x + 1


def diagram_to_nroff(nodes, para):
    pic_s = ''
    text = [''] * 7
    x = 0
    move = False
    for node in nodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName == 'header':
            if move:
                pic_s += "move .1\n"
                x += 1
            elif x > 0:
                x -= 1
            pic_header, x = diagram_header_to_nroff(node, text, x)
            pic_s += "[\n" + pic_header + "]\n"
            move = True
        elif node.nodeType == node.ELEMENT_NODE and node.tagName == 'nospace':
            move = False
        elif node.nodeType == node.ELEMENT_NODE and node.tagName == 'dots':
            pic_s += "move .1\n"
            pic_s += '". . ." ljust\n'

            put_text(text, x, 4, " ... ")
            x += 5
        elif node.nodeType == node.COMMENT_NODE:
            pass
        elif node.nodeType == node.TEXT_NODE and node.data.isspace():
            pass
        else:
            fatal("unknown node %s in diagram <header> element" % node)

    text_s = '.br\n'.join(["\\fL%s\n" % s for s in text if s != ""])
    return para + """
.\\" check if in troff mode (TTY)
.if t \\{
.PS
boxht = .2
textht = 1/6
fillval = .2
""" + pic_s + """\
.PE
\\}
.\\" check if in nroff mode:
.if n \\{
.nf
""" + text_s + """\
.fi
\\}"""


def flatten_header(s):
    s = s.strip()
    return re.sub(r'\s+', ' ', s)


def block_xml_to_nroff(nodes, para='.PP'):
    HEADER_TAGS = ('h1', 'h2', 'h3', 'h4')
    s = ''
    prev = ''
    for node in nodes:
        if node.nodeType == node.TEXT_NODE:
            if s == '' and para != '.IP':
                s = para + '\n'
            text = re.sub(r'\s+', ' ', node.data)
            if s.endswith(' '):
                text = text.lstrip()
            s += text_to_nroff(text)
            s = s.lstrip()
        elif node.nodeType == node.ELEMENT_NODE:
            if node.tagName in ['ul', 'ol']:
                if s != "":
                    s += "\n"
                s += ".RS\n"
                i = 0
                for li_node in node.childNodes:
                    if (li_node.nodeType == node.ELEMENT_NODE
                        and li_node.tagName == 'li'):
                        i += 1
                        if node.tagName == 'ul':
                            s += ".IP \\(bu\n"
                        else:
                            s += ".IP %d. .4in\n" % i
                        s += block_xml_to_nroff(li_node.childNodes, ".IP")
                    elif li_node.nodeType == node.COMMENT_NODE:
                        pass
                    elif (li_node.nodeType != node.TEXT_NODE
                          or not li_node.data.isspace()):
                        raise error.Error("<%s> element may only have "
                                          "<li> children" % node.tagName)
                s += ".RE\n"
            elif node.tagName == 'dl':
                indent = True
                if prev in HEADER_TAGS:
                    indent = False
                if s != "":
                    s += "\n"
                if indent:
                    s += ".RS\n"
                prev = "dd"
                for li_node in node.childNodes:
                    if (li_node.nodeType == node.ELEMENT_NODE
                        and li_node.tagName == 'dt'):
                        if prev == 'dd':
                            s += '.TP\n'
                        else:
                            s += '.TQ .5in\n'
                        prev = 'dt'
                    elif (li_node.nodeType == node.ELEMENT_NODE
                          and li_node.tagName == 'dd'):
                        if prev == 'dd':
                            s += '.IP\n'
                        prev = 'dd'
                    elif li_node.nodeType == node.COMMENT_NODE:
                        continue
                    elif (li_node.nodeType != node.TEXT_NODE
                          or not li_node.data.isspace()):
                        raise error.Error("<dl> element may only have "
                                          "<dt> and <dd> children")
                    s += block_xml_to_nroff(li_node.childNodes, ".IP")
                if indent:
                    s += ".RE\n"
            elif node.tagName == 'p':
                if s != "":
                    if not s.endswith("\n"):
                        s += "\n"
                    s += para + "\n"
                s += block_xml_to_nroff(node.childNodes, para)
            elif node.tagName in HEADER_TAGS:
                if s != "":
                    if not s.endswith("\n"):
                        s += "\n"
                nroffTag, font = {'h1': ('SH', r'\fR'),
                                  'h2': ('SS', r'\fB'),
                                  'h3': ('ST', r'\fI'),
                                  'h4': ('SU', r'\fI')}[node.tagName]
                to_upper = node.tagName == 'h1'
                s += ".%s \"" % nroffTag
                s += flatten_header(''.join([
                    inline_xml_to_nroff(child_node, font, to_upper)
                    for child_node in node.childNodes]))
                s += "\"\n"
            elif node.tagName == 'pre':
                fixed = node.getAttribute('fixed')
                if fixed == 'yes':
                    font = r'\fL'
                else:
                    font = r'\fB'
                s += pre_to_nroff(node.childNodes, para, font)
            elif node.tagName == 'tbl':
                s += tbl_to_nroff(node.childNodes, para)
            elif node.tagName == 'diagram':
                s += diagram_to_nroff(node.childNodes, para)
            else:
                s += inline_xml_to_nroff(node, r'\fR')
            prev = node.tagName
        elif node.nodeType == node.COMMENT_NODE:
            pass
        else:
            raise error.Error("unknown node %s in block xml" % node)
    if s != "" and not s.endswith('\n'):
        s += '\n'
    return s



#
# Copyright (c) 2010, 2012, 2013 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import signal
import socket
import sys

import ovs.socket_util
from ovs.fatal_signal import signal_alarm


def main(argv):
    if len(argv) not in (2, 3):
        sys.stderr.write("usage: %s SOCKETNAME1 [SOCKETNAME2]", argv[0])
        sys.exit(1)

    sockname1 = argv[1]
    if len(argv) > 2:
        sockname2 = argv[2]
    else:
        sockname2 = sockname1

    signal.signal(signal.SIGALRM, signal.SIG_DFL)
    signal_alarm(5)

    # Create a listening socket under name 'sockname1'.
    error, sock1 = ovs.socket_util.make_unix_socket(socket.SOCK_STREAM, False,
                                                    sockname1, None)
    if error:
        sys.stderr.write("%s: bind failed (%s)" % (sockname1,
                                                   os.strerror(error)))
        sys.exit(1)
    sock1.listen(1)

    # Connect to 'sockname2' (which should be the same file, perhaps under a
    # different name).
    error, sock2 = ovs.socket_util.make_unix_socket(socket.SOCK_STREAM, False,
                                                    None, sockname2)
    if error:
        sys.stderr.write("%s: connect failed (%s)" % (sockname2,
                                                      os.strerror(error)))
        sys.exit(1)


if __name__ == '__main__':
    main(sys.argv)



# Copyright (c) 2011, 2012 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
tcp module contains listener and sender classes for TCP protocol
"""

import time

from twisted.internet import interfaces
from twisted.internet.protocol import ClientFactory, Factory, Protocol

from zope.interface.declarations import implementer


class TcpListenerConnection(Protocol):
    """
    This per-connection class is instantiated each time sender connects
    """
    def __init__(self):
        self.stats = 0

    def dataReceived(self, data):
        self.stats += len(data)

    def connectionLost(self, reason):
        self.factory.stats += self.stats


class TcpListenerFactory(Factory):
    """
    This per-listening socket class is used to
    instantiate TcpListenerConnections
    """
    protocol = TcpListenerConnection

    def __init__(self):
        self.stats = 0

    def getResults(self):
        """ returns the number of bytes received as string"""
        # XML RPC does not support 64bit int (http://bugs.python.org/issue2985)
        # so we have to convert the amount of bytes into a string
        return str(self.stats)


@implementer(interfaces.IPushProducer)
class Producer(object):
    """
    This producer class generates infinite byte stream for a specified time
    duration
    """
    def __init__(self, proto, duration):
        self.proto = proto
        self.start = time.time()
        self.produced = 0
        self.paused = False
        self.data = "X" * 65535
        self.duration = duration

    def pauseProducing(self):
        """This function is called whenever write() to socket would block"""
        self.paused = True

    def resumeProducing(self):
        """This function is called whenever socket becomes writable"""
        self.paused = False
        current = time.time()
        while (not self.paused) and (current < self.start + self.duration):
            self.proto.transport.write(self.data)
            self.produced += len(self.data)
            current = time.time()
        if current >= self.start + self.duration:
            self.proto.factory.stats += self.produced
            self.proto.transport.unregisterProducer()
            self.proto.transport.loseConnection()

    def stopProducing(self):
        pass


class TcpSenderConnection(Protocol):
    """
    TCP connection instance class that sends all traffic at full speed.
    """

    def connectionMade(self):
        producer = Producer(self, self.factory.duration)
        self.transport.registerProducer(producer, True)
        producer.resumeProducing()

    def dataReceived(self, data):
        self.transport.loseConnection()


class TcpSenderFactory(ClientFactory):
    """
    This factory is responsible to instantiate TcpSenderConnection classes
    each time sender initiates connection
    """
    protocol = TcpSenderConnection

    def __init__(self, duration):
        self.duration = duration
        self.stats = 0

    def getResults(self):
        """Returns amount of bytes sent to the Listener (as a string)"""
        return str(self.stats)



# Copyright (c) 2012 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys

import ovs.daemon
import ovs.unixctl
import ovs.unixctl.client
import ovs.util
import ovs.vlog
from ovs.fatal_signal import signal_alarm


def connect_to_target(target):
    error, str_result = ovs.unixctl.socket_name_from_target(target)
    if error:
        ovs.util.ovs_fatal(error, str_result)
    else:
        socket_name = str_result

    error, client = ovs.unixctl.client.UnixctlClient.create(socket_name)
    if error:
        ovs.util.ovs_fatal(error, "cannot connect to \"%s\"" % socket_name)

    return client


def reply_to_string(reply, fmt=ovs.unixctl.UnixctlOutputFormat.TEXT,
                    fmt_flags={}):
    if fmt == ovs.unixctl.UnixctlOutputFormat.TEXT:
        body = str(reply)
    else:
        body = ovs.json.to_string(reply, **fmt_flags)

    if body and not body.endswith("\n"):
        body += "\n"

    return body


def main():
    parser = argparse.ArgumentParser(description="Python Implementation of"
                                     " ovs-appctl.")
    parser.add_argument("-t", "--target", default="ovs-vswitchd",
                        help="pidfile or socket to contact")

    parser.add_argument("command", metavar="COMMAND",
                        help="Command to run.")
    parser.add_argument("argv", metavar="ARG", nargs="*",
                        help="Arguments to the command.")
    parser.add_argument("-T", "--timeout", metavar="SECS",
                        help="wait at most SECS seconds for a response")
    parser.add_argument("-f", "--format", metavar="FMT",
                        help="Output format.", default="text",
                        choices=[fmt.name.lower()
                                 for fmt in ovs.unixctl.UnixctlOutputFormat],
                        type=str.lower)
    parser.add_argument("--pretty", action="store_true",
                        help="Format the output in a more readable fashion."
                             " Requires: --format json.")
    args = parser.parse_args()

    if (args.format != ovs.unixctl.UnixctlOutputFormat.JSON.name.lower()
        and args.pretty):
        ovs.util.ovs_fatal(0, "--pretty is supported with --format json only")

    signal_alarm(int(args.timeout) if args.timeout else None)

    ovs.vlog.Vlog.init()
    target = args.target
    format = ovs.unixctl.UnixctlOutputFormat[args.format.upper()]
    format_flags = dict(pretty=True) if args.pretty else {}
    client = connect_to_target(target)

    if format != ovs.unixctl.UnixctlOutputFormat.TEXT:
        err_no, error, _ = client.transact(
            "set-options", ["--format", args.format])

        if err_no:
            ovs.util.ovs_fatal(err_no, "%s: transaction error" % target)
        elif error is not None:
            sys.stderr.write(reply_to_string(error))
            ovs.util.ovs_error(0, "%s: server returned an error" % target)
            sys.exit(2)

    err_no, error, result = client.transact(args.command, args.argv)
    client.close()

    if err_no:
        ovs.util.ovs_fatal(err_no, "%s: transaction error" % target)
    elif error is not None:
        sys.stderr.write(reply_to_string(error))
        ovs.util.ovs_error(0, "%s: server returned an error" % target)
        sys.exit(2)
    else:
        assert result is not None
        sys.stdout.write(reply_to_string(result, format, format_flags))


if __name__ == '__main__':
    main()



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovs.flowviz.format import FlowFormatter, FlowBuffer, FlowStyle


class HTMLStyle:
    """HTMLStyle defines a style for html-formatted flows.

    Args:
        color(str): Optional; a string representing the CSS color to use
        anchor_gen(callable): Optional; a callable to be used to generate the
            href
    """

    def __init__(self, color=None, anchor_gen=None):
        self.color = color
        self.anchor_gen = anchor_gen


class HTMLBuffer(FlowBuffer):
    """HTMLBuffer implementes FlowBuffer to provide html-based flow formatting.

    Each flow gets formatted as:
        <div><span>...</span></div>
    """

    def __init__(self):
        self._text = ""

    @property
    def text(self):
        return self._text

    def _append(self, string, color, href):
        """Append a key a string"""
        style = ' style="color:{}"'.format(color) if color else ""
        self._text += "<span{}>".format(style)
        if href:
            self._text += "<a href={} {}> ".format(
                href, 'style="color:{}"'.format(color) if color else ""
            )
        self._text += string
        if href:
            self._text += "</a>"
        self._text += "</span>"

    def append_key(self, kv, style):
        """Append a key.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (HTMLStyle): the style to use
        """
        href = style.anchor_gen(kv) if (style and style.anchor_gen) else ""
        return self._append(
            kv.meta.kstring, style.color if style else "", href
        )

    def append_delim(self, kv, style):
        """Append a delimiter.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (HTMLStyle): the style to use
        """
        href = style.anchor_gen(kv) if (style and style.anchor_gen) else ""
        return self._append(kv.meta.delim, style.color if style else "", href)

    def append_end_delim(self, kv, style):
        """Append an end delimiter.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (HTMLStyle): the style to use
        """
        href = style.anchor_gen(kv) if (style and style.anchor_gen) else ""
        return self._append(
            kv.meta.end_delim, style.color if style else "", href
        )

    def append_value(self, kv, style):
        """Append a value.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (HTMLStyle): the style to use
        """
        href = style.anchor_gen(kv) if (style and style.anchor_gen) else ""
        return self._append(
            kv.meta.vstring, style.color if style else "", href
        )

    def append_value_omitted(self, kv):
        """Append an omitted value.
        Args:
            kv (KeyValue): the KeyValue instance to append
        """
        dots = "." * len(kv.meta.vstring)
        return self._append(dots, "", "")

    def append_extra(self, extra, style):
        """Append extra string.
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (HTMLStyle): the style to use
        """
        return self._append(extra, style.color if style else "", "")


class HTMLFormatter(FlowFormatter):
    """Formts a flow in HTML Format."""

    default_style_obj = FlowStyle(
        {
            "value.resubmit": HTMLStyle(
                anchor_gen=lambda x: "#table_{}".format(x.value["table"])
            ),
            "default": HTMLStyle(),
        }
    )

    def __init__(self, opts=None):
        super(HTMLFormatter, self).__init__()
        self.style = (
            self._style_from_opts(opts, "html", HTMLStyle) or FlowStyle()
        )

    def format_flow(self, buf, flow, highlighted=None, omitted=None):
        """Formats the flow into the provided buffer as a html object.

        Args:
            buf (FlowBuffer): the flow buffer to append to
            flow (ovs_dbg.OFPFlow): the flow to format
            highlighted (list): Optional; list of KeyValues to highlight
            omitted (list): Optional; list of KeyValues to omit
        """
        return super(HTMLFormatter, self).format_flow(
            buf, flow, self.style, highlighted, omitted
        )



#!/usr/bin/env python3
#
# Copyright (c) 2022 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Script information:
# -------------------
# dpif_nl_exec_monitor.py uses the dpif_netlink_operate__:op_flow_execute USDT
# probe to receive all DPIF_OP_EXECUTE operations that are queued for
# transmission over the netlink socket. It will do some basic decoding, and if
# requested a packet dump. Note that there are also options to obtain
# additional information for the DPIF_OP_FLOW_* operations.
#
# Here is an example:
#
#   # ./dpif_op_nl_monitor.py --packet-decode decode --trace-exec-op
#   Display DPIF_OP_EXECUTE operations being queued for transmission...
#   TIME               CPU  COMM             PID        NL_SIZE
#   Display DPIF operations being queued for transmission onto the netlink...
#   TIME               CPU  COMM             PID        NL_SIZE    DPIF_...
#   2162030.641541513  5    handler4         2951007    164        DPIF_...
#      nlmsghdr  : len = 0, type = 36, flags = 1, seq = 0, pid = 0
#      genlmsghdr: cmd = DPIF_OP_EXECUTE, version = 1, reserved = 0
#      ovs_header: dp_ifindex = 21
#        > Decode OVS_PACKET_ATTR_* TLVs:
#        nla_len 46, nla_type OVS_PACKET_ATTR_PACKET[1], data: 00 00 00...
#        nla_len 20, nla_type OVS_PACKET_ATTR_KEY[2], data: 08 00 02 00...
#            > Decode OVS_KEY_ATTR_* TLVs:
#            nla_len 8, nla_type OVS_KEY_ATTR_PRIORITY[2], data: 00 00...
#            nla_len 8, nla_type OVS_KEY_ATTR_SKB_MARK[15], data: 00 00...
#        nla_len 88, nla_type OVS_PACKET_ATTR_ACTIONS[3], data: 4c 00 03...
#            > Decode OVS_ACTION_ATTR_* TLVs:
#            nla_len 76, nla_type OVS_ACTION_ATTR_SET[3], data: 48 00...
#                    > Decode OVS_TUNNEL_KEY_ATTR_* TLVs:
#                    nla_len 12, nla_type OVS_TUNNEL_KEY_ATTR_ID[0], data:...
#                    nla_len 20, nla_type OVS_TUNNEL_KEY_ATTR_IPV6_DST[13], ...
#                    nla_len 5, nla_type OVS_TUNNEL_KEY_ATTR_TTL[4], data: 40
#                    nla_len 4, nla_type OVS_TUNNEL_KEY_ATTR_DONT_FRAGMENT...
#                    nla_len 4, nla_type OVS_TUNNEL_KEY_ATTR_CSUM[6], data:
#                    nla_len 6, nla_type OVS_TUNNEL_KEY_ATTR_TP_DST[10],...
#                    nla_len 12, nla_type OVS_TUNNEL_KEY_ATTR_GENEVE_OPTS...
#            nla_len 8, nla_type OVS_ACTION_ATTR_OUTPUT[1], data: 02 00 00 00
#        - Dumping OVS_PACKET_ATR_PACKET data:
#        ###[ Ethernet ]###
#          dst       = 00:00:00:00:ec:01
#          src       = 04:f4:bc:28:57:00
#          type      = IPv4
#        ###[ IP ]###
#             version   = 4
#             ihl       = 5
#             tos       = 0x0
#             len       = 50
#             id        = 0
#             flags     =
#             frag      = 0
#             ttl       = 127
#             proto     = icmp
#             chksum    = 0x2767
#             src       = 10.0.0.1
#             dst       = 10.0.0.100
#             \options   \
#        ###[ ICMP ]###
#                type      = echo-request
#                code      = 0
#                chksum    = 0xf7f3
#                id        = 0x0
#                seq       = 0xc
#
# The example above dumps the full netlink and packet decode. However options
# exist to disable this. For a complete list of options, please use the
# '--help' or '-h' argument.
#

from bcc import BPF, USDT, USDTException
from os.path import exists
from scapy.all import hexdump, wrpcap
from scapy.layers.l2 import Ether

import argparse
import psutil
import re
import struct
import sys
import time

#
# Actual eBPF source code
#
ebpf_source = """
#include <linux/sched.h>

#define MAX_NLMSG <MAX_NLMSG_VAL>

struct event_t {
    u32 cpu;
    u32 pid;
    u64 ts;
    u32 nl_size;
    char comm[TASK_COMM_LEN];
    u8 nl_msg[MAX_NLMSG];
};

struct ofpbuf {
    void *base;
    void *data;
    uint32_t size;

    /* The actual structure is longer, but we are only interested in the
     * first couple of entries. */
};

BPF_RINGBUF_OUTPUT(events, <BUFFER_PAGE_CNT>);
BPF_TABLE("percpu_array", uint32_t, uint64_t, dropcnt, 1);

static int trace_event(struct ofpbuf *nlbuf)
{
    uint32_t size;

    struct event_t *event = events.ringbuf_reserve(sizeof(struct event_t));
    if (!event) {
        uint32_t type = 0;
        uint64_t *value = dropcnt.lookup(&type);
        if (value)
            __sync_fetch_and_add(value, 1);

        return 1;
    }

    event->ts = bpf_ktime_get_ns();
    event->cpu =  bpf_get_smp_processor_id();
    event->pid = bpf_get_current_pid_tgid();
    bpf_get_current_comm(&event->comm, sizeof(event->comm));

    event->nl_size = nlbuf->size;
    if (event->nl_size > MAX_NLMSG)
        size = MAX_NLMSG;
    else
        size = event->nl_size;

    bpf_probe_read(&event->nl_msg, size, nlbuf->data);

    events.ringbuf_submit(event, 0);
    return 0;
}

#if <ENABLE_OP_EXECUTE>
int trace__op_execute(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(5, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif

#if <ENABLE_OP_FLOW_PUT>
int trace__op_flow_put(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(4, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif

#if <ENABLE_OP_FLOW_DEL>
int trace__op_flow_del(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(4, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif

#if <ENABLE_OP_FLOW_GET>
int trace__op_flow_get(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(4, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif
"""


#
# print_event()
#
def print_event(ctx, data, size):
    event = b["events"].event(data)

    if event.nl_size < options.nlmsg_size:
        nl_size = event.nl_size
    else:
        nl_size = options.nlmsg_size

    print("{:<18.9f} {:<4} {:<16} {:<10} {:<10} {}".
          format(event.ts / 1000000000,
                 event.cpu,
                 event.comm.decode("utf-8"),
                 event.pid,
                 event.nl_size,
                 get_ovs_dpif_op_str(get_cmd_type_from_nlm(
                     bytes(event.nl_msg)[:nl_size]))))

    #
    # Dumping the netlink message data if requested.
    #
    if options.nlmsg_decode == "hex":
        #
        # Abuse scapy's hex dump to dump flow key
        #
        print(re.sub("^", " " * 4,
                     hexdump(Ether(bytes(event.nl_msg)[:nl_size]), dump=True),
                     flags=re.MULTILINE))

    if options.nlmsg_decode == "nlraw":
        decode_result = decode_nlm(bytes(event.nl_msg)[:nl_size], dump=True)
    else:
        decode_result = decode_nlm(bytes(event.nl_msg)[:nl_size], dump=False)

    #
    # Decode packet only if there is data
    #
    if "OVS_PACKET_ATTR_PACKET" not in decode_result:
        return

    pkt_data = decode_result["OVS_PACKET_ATTR_PACKET"]
    indent = 4 if options.nlmsg_decode != "nlraw" else 6

    if options.packet_decode != "none":
        print("{}- Dumping OVS_PACKET_ATR_PACKET data:".format(" " * indent))

    if options.packet_decode == "hex":
        print(re.sub("^", " " * indent, hexdump(pkt_data, dump=True),
                     flags=re.MULTILINE))

    packet = Ether(pkt_data)
    if options.packet_decode == "decode":
        print(re.sub("^", " " * indent, packet.show(dump=True),
                     flags=re.MULTILINE))

    if options.pcap is not None:
        wrpcap(options.pcap, packet, append=True)


#
# decode_nlm_tlvs()
#
def decode_nlm_tlvs(tlvs, header=None, indent=4, dump=True,
                    attr_to_str_func=None, decode_tree=None):
    bytes_left = len(tlvs)
    result = {}

    if dump:
        print("{}{}".format(" " * indent, header))

    while bytes_left:
        if bytes_left < 4:
            if dump:
                print("{}WARN: decode truncated; can't read header".format(
                    " " * indent))
            break

        nla_len, nla_type = struct.unpack("=HH", tlvs[:4])

        if nla_len < 4:
            if dump:
                print("{}WARN: decode truncated; nla_len < 4".format(
                    " " * indent))
            break

        nla_data = tlvs[4:nla_len]
        trunc = ""

        if attr_to_str_func is None:
            nla_type_name = "type_{}".format(nla_type)
        else:
            nla_type_name = attr_to_str_func(nla_type)

        if nla_len > bytes_left:
            trunc = "..."
            nla_data = nla_data[:(bytes_left - 4)]
        else:
            result[nla_type_name] = nla_data

        if dump:
            print("{}nla_len {}, nla_type {}[{}], data: {}{}".format(
                " " * indent, nla_len, nla_type_name, nla_type,
                "".join("{:02x} ".format(b) for b in nla_data), trunc))

            #
            # If we have the full data, try to decode further
            #
            if trunc == "" and decode_tree is not None \
               and nla_type_name in decode_tree:
                node = decode_tree[nla_type_name]
                decode_nlm_tlvs(nla_data,
                                header=node["header"],
                                indent=indent + node["indent"], dump=True,
                                attr_to_str_func=node["attr_str_func"],
                                decode_tree=node["decode_tree"])

        if trunc != "":
            if dump:
                print("{}WARN: decode truncated; nla_len > msg_len[{}] ".
                      format(" " * indent, bytes_left))
            break

        # update next offset, but make sure it's aligned correctly
        next_offset = (nla_len + 3) & ~(3)
        tlvs = tlvs[next_offset:]
        bytes_left -= next_offset

    return result


#
# get_cmd_type_from_nlm()
#
def get_cmd_type_from_nlm(nlm):
    # The netlink message consists of at least a 'struct nlmsghdr' (16-bytes)
    # followed by a 'struct genlmsghdr'. The first byte of the genlmsghdr
    # structure contains the command. Which is what we will extract here.
    if len(nlm) < 17:
        return -1

    return nlm[16]


#
# decode_nlm()
#
def decode_nlm(msg, indent=4, dump=True):
    result = {}

    #
    # Decode 'struct nlmsghdr'
    #
    if dump:
        print("{}nlmsghdr  : len = {}, type = {}, flags = {}, seq = {}, "
              "pid = {}".format(" " * indent,
                                *struct.unpack("=IHHII", msg[:16])))

    msg = msg[16:]

    #
    # Decode 'struct genlmsghdr'
    #
    cmd, version, reserved = struct.unpack("=BBH", msg[:4])

    if dump:
        print("{}genlmsghdr: cmd = {}, version = {}, reserved = {}".format(
            " " * indent, get_ovs_dpif_op_str(cmd), version, reserved))

    msg = msg[4:]

    #
    # Decode 'struct ovs_header'
    #
    if dump:
        print("{}ovs_header: dp_ifindex = {}".format(
            " " * indent, *struct.unpack("=I", msg[:4])))

    msg = msg[4:]

    #
    # Decode TLVs
    #
    nl_key_attr = {
        "header": "> Decode OVS_KEY_ATTR_* TLVs:",
        "indent": 4,
        "attr_str_func": get_ovs_key_attr_str,
        "decode_tree": {
            "OVS_KEY_ATTR_ENCAP": {
                "header": "> Decode OVS_KEY_ATTR_* TLVs:",
                "indent": 4,
                "attr_str_func": get_ovs_key_attr_str,
                "decode_tree": {
                    "OVS_KEY_ATTR_ENCAP": {
                        "header": "> Decode OVS_KEY_ATTR_* TLVs:",
                        "indent": 4,
                        "attr_str_func": get_ovs_key_attr_str,
                        "decode_tree": None,
                    },
                },
            },
        }
    }

    nl_action_attr = {
        "header": "> Decode OVS_ACTION_ATTR_* TLVs:",
        "indent": 4,
        "attr_str_func": get_ovs_action_attr_str,
        "decode_tree": {
            "OVS_ACTION_ATTR_SET": {
                "header": "> Decode OVS_KEY_ATTR_* TLVs:",
                "indent": 4,
                "attr_str_func": get_ovs_key_attr_str,
                "decode_tree": {
                    "OVS_KEY_ATTR_TUNNEL": {
                        "header": "> Decode OVS_TUNNEL_KEY_ATTR_* TLVs:",
                        "indent": 4,
                        "attr_str_func": get_ovs_tunnel_key_attr_str,
                        "decode_tree": None,
                    },
                },
            },
        },
    }

    nl_attr_tree_exec = {
        "OVS_PACKET_ATTR_KEY": nl_key_attr,
        "OVS_PACKET_ATTR_ACTIONS": nl_action_attr,
    }

    nl_attr_tree_put = {
        "OVS_FLOW_ATTR_KEY": nl_key_attr,
        "OVS_FLOW_ATTR_MASK": nl_key_attr,
        "OVS_FLOW_ATTR_ACTIONS": nl_action_attr,
    }

    if get_ovs_dpif_op_str(cmd) == "DPIF_OP_EXECUTE":
        result = decode_nlm_tlvs(msg, indent=indent + 2, dump=dump,
                                 header="> Decode OVS_PACKET_ATTR_* TLVs:",
                                 attr_to_str_func=get_ovs_pkt_attr_str,
                                 decode_tree=nl_attr_tree_exec)
    else:
        result = decode_nlm_tlvs(msg, indent=indent + 2, dump=dump,
                                 header="> Decode OVS_FLOW_ATTR_* TLVs:",
                                 attr_to_str_func=get_ovs_flow_attr_str,
                                 decode_tree=nl_attr_tree_put)
    return result


#
# get_ovs_flow_attr_str()
#
def get_ovs_flow_attr_str(attr):
    ovs_flow_attr = ["OVS_FLOW_ATTR_UNSPEC",
                     "OVS_FLOW_ATTR_KEY",
                     "OVS_FLOW_ATTR_ACTIONS",
                     "OVS_FLOW_ATTR_STATS",
                     "OVS_FLOW_ATTR_TCP_FLAGS",
                     "OVS_FLOW_ATTR_USED",
                     "OVS_FLOW_ATTR_CLEAR",
                     "OVS_FLOW_ATTR_MASK",
                     "OVS_FLOW_ATTR_PROBE",
                     "OVS_FLOW_ATTR_UFID",
                     "OVS_FLOW_ATTR_UFID_FLAGS",
                     "OVS_FLOW_ATTR_PAD"]
    if attr < 0 or attr >= len(ovs_flow_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_flow_attr[attr]


#
# get_ovs_pkt_attr_str()
#
def get_ovs_pkt_attr_str(attr):
    ovs_pkt_attr = ["OVS_PACKET_ATTR_UNSPEC",
                    "OVS_PACKET_ATTR_PACKET",
                    "OVS_PACKET_ATTR_KEY",
                    "OVS_PACKET_ATTR_ACTIONS",
                    "OVS_PACKET_ATTR_USERDATA",
                    "OVS_PACKET_ATTR_EGRESS_TUN_KEY",
                    "OVS_PACKET_ATTR_UNUSED1",
                    "OVS_PACKET_ATTR_UNUSED2",
                    "OVS_PACKET_ATTR_PROBE",
                    "OVS_PACKET_ATTR_MRU",
                    "OVS_PACKET_ATTR_LEN",
                    "OVS_PACKET_ATTR_HASH"]
    if attr < 0 or attr >= len(ovs_pkt_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_pkt_attr[attr]


#
# get_ovs_key_attr_str()
#
def get_ovs_key_attr_str(attr):
    ovs_key_attr = ["OVS_KEY_ATTR_UNSPEC",
                    "OVS_KEY_ATTR_ENCAP",
                    "OVS_KEY_ATTR_PRIORITY",
                    "OVS_KEY_ATTR_IN_PORT",
                    "OVS_KEY_ATTR_ETHERNET",
                    "OVS_KEY_ATTR_VLAN",
                    "OVS_KEY_ATTR_ETHERTYPE",
                    "OVS_KEY_ATTR_IPV4",
                    "OVS_KEY_ATTR_IPV6",
                    "OVS_KEY_ATTR_TCP",
                    "OVS_KEY_ATTR_UDP",
                    "OVS_KEY_ATTR_ICMP",
                    "OVS_KEY_ATTR_ICMPV6",
                    "OVS_KEY_ATTR_ARP",
                    "OVS_KEY_ATTR_ND",
                    "OVS_KEY_ATTR_SKB_MARK",
                    "OVS_KEY_ATTR_TUNNEL",
                    "OVS_KEY_ATTR_SCTP",
                    "OVS_KEY_ATTR_TCP_FLAGS",
                    "OVS_KEY_ATTR_DP_HASH",
                    "OVS_KEY_ATTR_RECIRC_ID",
                    "OVS_KEY_ATTR_MPLS",
                    "OVS_KEY_ATTR_CT_STATE",
                    "OVS_KEY_ATTR_CT_ZONE",
                    "OVS_KEY_ATTR_CT_MARK",
                    "OVS_KEY_ATTR_CT_LABELS",
                    "OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV4",
                    "OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV6",
                    "OVS_KEY_ATTR_NSH"]

    if attr < 0 or attr >= len(ovs_key_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_key_attr[attr]


#
# get_ovs_action_attr_str()
#
def get_ovs_action_attr_str(attr):
    ovs_action_attr = ["OVS_ACTION_ATTR_UNSPEC",
                       "OVS_ACTION_ATTR_OUTPUT",
                       "OVS_ACTION_ATTR_USERSPACE",
                       "OVS_ACTION_ATTR_SET",
                       "OVS_ACTION_ATTR_PUSH_VLAN",
                       "OVS_ACTION_ATTR_POP_VLAN",
                       "OVS_ACTION_ATTR_SAMPLE",
                       "OVS_ACTION_ATTR_RECIRC",
                       "OVS_ACTION_ATTR_HASH",
                       "OVS_ACTION_ATTR_PUSH_MPLS",
                       "OVS_ACTION_ATTR_POP_MPLS",
                       "OVS_ACTION_ATTR_SET_MASKED",
                       "OVS_ACTION_ATTR_CT",
                       "OVS_ACTION_ATTR_TRUNC",
                       "OVS_ACTION_ATTR_PUSH_ETH",
                       "OVS_ACTION_ATTR_POP_ETH",
                       "OVS_ACTION_ATTR_CT_CLEAR",
                       "OVS_ACTION_ATTR_PUSH_NSH",
                       "OVS_ACTION_ATTR_POP_NSH",
                       "OVS_ACTION_ATTR_METER",
                       "OVS_ACTION_ATTR_CLONE",
                       "OVS_ACTION_ATTR_CHECK_PKT_LEN",
                       "OVS_ACTION_ATTR_ADD_MPLS",
                       "OVS_ACTION_ATTR_DEC_TTL",
                       "OVS_ACTION_ATTR_DROP",
                       "OVS_ACTION_ATTR_PSAMPLE",
                       "OVS_ACTION_ATTR_TUNNEL_PUSH",
                       "OVS_ACTION_ATTR_TUNNEL_POP",
                       "OVS_ACTION_ATTR_LB_OUTPUT"]
    if attr < 0 or attr >= len(ovs_action_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_action_attr[attr]


#
# get_ovs_tunnel_key_attr_str()
#
def get_ovs_tunnel_key_attr_str(attr):
    ovs_tunnel_key_attr = ["OVS_TUNNEL_KEY_ATTR_ID",
                           "OVS_TUNNEL_KEY_ATTR_IPV4_SRC",
                           "OVS_TUNNEL_KEY_ATTR_IPV4_DST",
                           "OVS_TUNNEL_KEY_ATTR_TOS",
                           "OVS_TUNNEL_KEY_ATTR_TTL",
                           "OVS_TUNNEL_KEY_ATTR_DONT_FRAGMENT",
                           "OVS_TUNNEL_KEY_ATTR_CSUM",
                           "OVS_TUNNEL_KEY_ATTR_OAM",
                           "OVS_TUNNEL_KEY_ATTR_GENEVE_OPTS",
                           "OVS_TUNNEL_KEY_ATTR_TP_SRC",
                           "OVS_TUNNEL_KEY_ATTR_TP_DST",
                           "OVS_TUNNEL_KEY_ATTR_VXLAN_OPTS",
                           "OVS_TUNNEL_KEY_ATTR_IPV6_SRC",
                           "OVS_TUNNEL_KEY_ATTR_IPV6_DST",
                           "OVS_TUNNEL_KEY_ATTR_PAD",
                           "OVS_TUNNEL_KEY_ATTR_ERSPAN_OPTS",
                           "OVS_TUNNEL_KEY_ATTR_GTPU_OPTS"]
    if attr < 0 or attr >= len(ovs_tunnel_key_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_tunnel_key_attr[attr]


#
# get_ovs_dpif_op_str()
#
def get_ovs_dpif_op_str(op):
    ovs_dpif_ops = ["DPIF_OP_UNSPEC",
                    "DPIF_OP_FLOW_PUT",
                    "DPIF_OP_FLOW_DEL",
                    "DPIF_OP_EXECUTE",
                    "DPIF_OP_FLOW_GET"]

    if op < 0 or op >= len(ovs_dpif_ops):
        return "<UNKNOWN:{}>".format(op)

    return ovs_dpif_ops[op]


#
# buffer_size_type()
#
def buffer_size_type(astr, min=64, max=2048):
    value = int(astr)
    if min <= value <= max:
        return value
    else:
        raise argparse.ArgumentTypeError(
            "value not in range {}-{}".format(min, max))


#
# next_power_of_two()
#
def next_power_of_two(val):
    np = 1
    while np < val:
        np *= 2
    return np


#
# main()
#
def main():
    #
    # Don't like these globals, but ctx passing does not seem to work with the
    # existing open_ring_buffer() API :(
    #
    global b
    global options

    #
    # Argument parsing
    #
    parser = argparse.ArgumentParser()

    parser.add_argument("--buffer-page-count",
                        help="Number of BPF ring buffer pages, default 1024",
                        type=int, default=1024, metavar="NUMBER")
    parser.add_argument("-D", "--debug",
                        help="Enable eBPF debugging",
                        type=int, const=0x3f, default=0, nargs="?")
    parser.add_argument("-d", "--packet-decode",
                        help="Display packet content in selected mode, "
                        "default none",
                        choices=["none", "hex", "decode"], default="none")
    parser.add_argument("-n", "--nlmsg-decode",
                        help="Display netlink message content in selected mode"
                        ", default nlraw",
                        choices=["none", "hex", "nlraw"], default="nlraw")
    parser.add_argument("-p", "--pid", metavar="VSWITCHD_PID",
                        help="ovs-vswitch's PID",
                        type=int, default=None)
    parser.add_argument("-s", "--nlmsg-size",
                        help="Set maximum netlink message size to capture, "
                        "default 512", type=buffer_size_type, default=512,
                        metavar="[64-2048]")
    parser.add_argument("--trace-del-op",
                        help="Monitor DPIF_OP_FLOW_DEL messages",
                        action="store_true")
    parser.add_argument("--trace-exec-op",
                        help="Monitor DPIF_OP_EXECUTE messages",
                        action="store_true")
    parser.add_argument("--trace-get-op",
                        help="Monitor DPIF_OP_FLOW_GET messages",
                        action="store_true")
    parser.add_argument("--trace-put-op",
                        help="Monitor DPIF_OP_FLOW_PUT messages",
                        action="store_true")
    parser.add_argument("-w", "--pcap", metavar="PCAP_FILE",
                        help="Write execute packets to specified pcap file",
                        type=str, default=None)

    options = parser.parse_args()

    #
    # Verify arguments.
    #
    if (not options.trace_del_op and not options.trace_exec_op
       and not options.trace_get_op and not options.trace_put_op):
        print("ERROR: At least on of the '--trace-*-op' arguments should be "
              "specified!")
        sys.exit(-1)

    #
    # Find the PID of the ovs-vswitchd daemon if not specified.
    #
    if options.pid is None:
        for proc in psutil.process_iter():
            if "ovs-vswitchd" in proc.name():
                if options.pid is not None:
                    print("ERROR: Multiple ovs-vswitchd daemons running, "
                          "use the -p option!")
                    sys.exit(-1)

                options.pid = proc.pid

    #
    # Error checking on input parameters
    #
    if options.pid is None:
        print("ERROR: Failed to find ovs-vswitchd's PID!")
        sys.exit(-1)

    if options.pcap is not None:
        if exists(options.pcap):
            print("ERROR: Destination capture file \"{}\" already exists!".
                  format(options.pcap))
            sys.exit(-1)

    options.buffer_page_count = next_power_of_two(options.buffer_page_count)

    #
    # Attach the usdt probe
    #
    u = USDT(pid=int(options.pid))
    try:
        if options.trace_exec_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_execute",
                           fn_name="trace__op_execute")
        if options.trace_put_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_put",
                           fn_name="trace__op_flow_put")
        if options.trace_del_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_del",
                           fn_name="trace__op_flow_del")
        if options.trace_get_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_get",
                           fn_name="trace__op_flow_get")
    except USDTException as e:
        print("ERROR: {}".format(
            (re.sub("^", " " * 7, str(e), flags=re.MULTILINE)).strip().
            replace("--with-dtrace or --enable-dtrace",
                    "--enable-usdt-probes")))
        sys.exit(-1)

    #
    # Uncomment to see how arguments are decoded.
    #   print(u.get_text())
    #

    #
    # Attach probe to running process
    #
    source = ebpf_source.replace("<MAX_NLMSG_VAL>", str(options.nlmsg_size))
    source = source.replace("<BUFFER_PAGE_CNT>",
                            str(options.buffer_page_count))

    source = source.replace("<ENABLE_OP_FLOW_DEL>",
                            "1" if options.trace_del_op else "0")
    source = source.replace("<ENABLE_OP_EXECUTE>",
                            "1" if options.trace_exec_op else "0")
    source = source.replace("<ENABLE_OP_FLOW_GET>",
                            "1" if options.trace_get_op else "0")
    source = source.replace("<ENABLE_OP_FLOW_PUT>",
                            "1" if options.trace_put_op else "0")

    b = BPF(text=source, usdt_contexts=[u], debug=options.debug)

    #
    # Print header
    #
    print("Display DPIF operations being queued for transmission onto the "
          "netlink socket.")
    print("{:<18} {:<4} {:<16} {:<10} {:<10} {}".format(
        "TIME", "CPU", "COMM", "PID", "NL_SIZE", "DPIF_OPERATION"))

    #
    # Dump out all events
    #
    b["events"].open_ring_buffer(print_event)
    while 1:
        try:
            b.ring_buffer_poll()
            time.sleep(0.5)
        except KeyboardInterrupt:
            break

    dropcnt = b.get_table("dropcnt")
    for k in dropcnt.keys():
        count = dropcnt.sum(k).value
        if k.value == 0 and count > 0:
            print("\nWARNING: Not all upcalls were captured, {} were dropped!"
                  "\n         Increase the BPF ring buffer size with the "
                  "--buffer-page-count option.".format(count))


#
# Start main() as the default entry point...
#
if __name__ == "__main__":
    main()



#! /usr/bin/env python3

# Copyright (c) 2008, 2017 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from ovs_build_helpers import soutil


def soexpand(include_dirs, src, dst):
    ok = True
    while True:
        line = src.readline()
        if not line:
            break

        name = soutil.extract_include_directive(line)
        if name:
            fn = soutil.find_file(include_dirs, name)
            if fn:
                try:
                    f = open(fn)
                    while True:
                        inner = f.readline()
                        if not inner:
                            break
                        dst.write(inner)
                except IOError as e:
                    sys.stderr.write("%s: open: %s\n" % (fn, e.strerror))
                    ok = False
                continue
            else:
                ok = False

        dst.write(line)
    return ok


if __name__ == '__main__':
    include_dirs, args = soutil.parse_include_dirs()
    if args:
        error = False
        for arg in args:
            if not soexpand(include_dirs, open(arg), sys.stdout):
                error = True
    else:
        error = not soexpand(include_dirs, sys.stdin, sys.stdout)
    sys.exit(1 if error else 0)



import pytest

from ovs.flow.list import ListParser, ListDecoders
from ovs.flow.kv import KeyValue


@pytest.mark.parametrize(
    "input_data,expected",
    [
        (
            ("field1,field2,3,nested:value", None, [","]),
            [
                KeyValue("elem_0", "field1"),
                KeyValue("elem_1", "field2"),
                KeyValue("elem_2", 3),
                KeyValue("elem_3", "nested:value"),
            ],
        ),
        (
            (
                "field1,field2,3,nested:value",
                ListDecoders(
                    [
                        ("key1", str),
                        ("key2", str),
                        ("key3", int),
                        ("key4", lambda x: x.split(":"), [","]),
                    ]
                ),
                [","],
            ),
            [
                KeyValue("key1", "field1"),
                KeyValue("key2", "field2"),
                KeyValue("key3", 3),
                KeyValue("key4", ["nested", "value"]),
            ],
        ),
        (
            ("field1:field2:3", None, [":"]),
            [
                KeyValue("elem_0", "field1"),
                KeyValue("elem_1", "field2"),
                KeyValue("elem_2", 3),
            ],
        ),
    ],
)
def test_kv_parser(input_data, expected):
    input_string = input_data[0]
    decoders = input_data[1]
    delims = input_data[2]
    tparser = ListParser(input_string, decoders, delims)
    tparser.parse()
    result = tparser.kv()
    assert len(expected) == len(result)
    for i in range(0, len(result)):
        assert result[i].key == expected[i].key
        assert result[i].value == expected[i].value
        kpos = result[i].meta.kpos
        kstr = result[i].meta.kstring
        vpos = result[i].meta.vpos
        vstr = result[i].meta.vstring
        assert input_string[kpos : kpos + len(kstr)] == kstr
        if vpos != -1:
            assert input_string[vpos : vpos + len(vstr)] == vstr



# Copyright (c) 2010, 2011 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import signal
import sys
import time

import ovs.daemon
import ovs.util


def handler(signum, _):
    raise Exception("Signal handler called with %d" % signum)


def main():

    if sys.platform != 'win32':
        # signal.SIGHUP does not exist on Windows
        signal.signal(signal.SIGHUP, handler)

    parser = argparse.ArgumentParser(
            description="Open vSwitch daemonization test program for Python.")
    parser.add_argument("-b", "--bail", action="store_true",
            help="Exit with an error after daemonize_start().")

    ovs.daemon.add_args(parser)
    args = parser.parse_args()
    ovs.daemon.handle_args(args)

    ovs.daemon.daemonize_start()
    if args.bail:
        sys.stderr.write("%s: exiting after daemonize_start() as requested\n"
                         % ovs.util.PROGRAM_NAME)
        sys.exit(1)
    ovs.daemon.daemonize_complete()

    while True:
        time.sleep(1)


if __name__ == '__main__':
    try:
        main()
    except SystemExit:
        # Let system.exit() calls complete normally
        raise
    except:
        sys.exit(ovs.daemon.RESTART_EXIT_CODE)



# Copyright (c) 2009, 2010, 2011 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import errno
import os
import sys

import ovs.daemon
import ovs.json
import ovs.jsonrpc
import ovs.poller
import ovs.stream


def handle_rpc(rpc, msg):
    done = False
    reply = None

    if msg.type == ovs.jsonrpc.Message.T_REQUEST:
        if msg.method == "echo":
            reply = ovs.jsonrpc.Message.create_reply(msg.params, msg.id)
        else:
            reply = ovs.jsonrpc.Message.create_error(
                {"error": "unknown method"}, msg.id)
            sys.stderr.write("unknown request %s" % msg.method)
    elif msg.type == ovs.jsonrpc.Message.T_NOTIFY:
        if msg.method == "shutdown":
            done = True
        else:
            rpc.error(errno.ENOTTY)
            sys.stderr.write("unknown notification %s" % msg.method)
    else:
        rpc.error(errno.EPROTO)
        sys.stderr.write("unsolicited JSON-RPC reply or error\n")

    if reply:
        rpc.send(reply)
    return done


def do_listen(name):
    if sys.platform != 'win32' or (
            ovs.daemon._detach and ovs.daemon._detached):
        # On Windows the child is a new process created which should be the
        # one that creates the PassiveStream. Without this check, the new
        # child process will create a new PassiveStream overwriting the one
        # that the parent process created.
        error, pstream = ovs.stream.PassiveStream.open(name)
        if error:
            sys.stderr.write("could not listen on \"%s\": %s\n"
                             % (name, os.strerror(error)))
            sys.exit(1)

    ovs.daemon.daemonize()

    rpcs = []
    done = False
    while True:
        # Accept new connections.
        error, stream = pstream.accept()
        if stream:
            rpcs.append(ovs.jsonrpc.Connection(stream))
        elif error != errno.EAGAIN:
            sys.stderr.write("PassiveStream.accept() failed\n")
            sys.exit(1)

        # Service existing connections.
        dead_rpcs = []
        for rpc in rpcs:
            rpc.run()

            error = 0
            if not rpc.get_backlog():
                error, msg = rpc.recv()
                if not error:
                    if handle_rpc(rpc, msg):
                        done = True

            error = rpc.get_status()
            if error:
                rpc.close()
                dead_rpcs.append(rpc)
        rpcs = [rpc for rpc in rpcs if rpc not in dead_rpcs]

        if done and not rpcs:
            break

        poller = ovs.poller.Poller()
        pstream.wait(poller)
        for rpc in rpcs:
            rpc.wait(poller)
            if not rpc.get_backlog():
                rpc.recv_wait(poller)
        poller.block()
    pstream.close()


def do_request(name, method, params_string):
    params = ovs.json.from_string(params_string)
    msg = ovs.jsonrpc.Message.create_request(method, params)
    s = msg.is_valid()
    if s:
        sys.stderr.write("not a valid JSON-RPC request: %s\n" % s)
        sys.exit(1)

    error, stream = ovs.stream.Stream.open_block(ovs.stream.Stream.open(name))
    if error:
        sys.stderr.write("could not open \"%s\": %s\n"
                         % (name, os.strerror(error)))
        sys.exit(1)

    rpc = ovs.jsonrpc.Connection(stream)

    error = rpc.send(msg)
    if error:
        sys.stderr.write("could not send request: %s\n" % os.strerror(error))
        sys.exit(1)

    error, msg = rpc.recv_block()
    if error:
        sys.stderr.write("error waiting for reply: %s\n" % os.strerror(error))
        sys.exit(1)

    print(ovs.json.to_string(msg.to_json()))

    rpc.close()


def do_notify(name, method, params_string):
    params = ovs.json.from_string(params_string)
    msg = ovs.jsonrpc.Message.create_notify(method, params)
    s = msg.is_valid()
    if s:
        sys.stderr.write("not a valid JSON-RPC notification: %s\n" % s)
        sys.exit(1)

    error, stream = ovs.stream.Stream.open_block(ovs.stream.Stream.open(name))
    if error:
        sys.stderr.write("could not open \"%s\": %s\n"
                         % (name, os.strerror(error)))
        sys.exit(1)

    rpc = ovs.jsonrpc.Connection(stream)

    error = rpc.send_block(msg)
    if error:
        sys.stderr.write("could not send notification: %s\n"
                         % os.strerror(error))
        sys.exit(1)

    rpc.close()


def main(argv):

    parser = argparse.ArgumentParser(
            description="JSON-RPC test utility for Python.",
            formatter_class=argparse.RawDescriptionHelpFormatter)

    commands = {"listen": (do_listen, 1),
                "request": (do_request, 3),
                "notify": (do_notify, 3),
                "help": (parser.print_help, (0,))}

    group_description = """\
listen LOCAL             listen for connections on LOCAL
request REMOTE METHOD PARAMS   send request, print reply
notify REMOTE METHOD PARAMS  send notification and exit
""" + ovs.stream.usage("JSON-RPC")

    group = parser.add_argument_group(title="Commands",
                                      description=group_description)
    group.add_argument('command', metavar="COMMAND", nargs=1,
                        choices=commands, help="Command to use.")
    group.add_argument('command_args', metavar="ARG", nargs='*',
                       help="Arguments to COMMAND.")

    ovs.daemon.add_args(parser)
    args = parser.parse_args()
    ovs.daemon.handle_args(args)

    command_name = args.command[0]
    args = args.command_args
    if command_name not in commands:
        sys.stderr.write("%s: unknown command \"%s\" "
                         "(use --help for help)\n" % (argv[0], command_name))
        sys.exit(1)

    func, n_args = commands[command_name]
    if type(n_args) is tuple:
        if len(args) < n_args[0]:
            sys.stderr.write("%s: \"%s\" requires at least %d arguments but "
                             "only %d provided\n"
                             % (argv[0], command_name, n_args, len(args)))
            sys.exit(1)
    elif type(n_args) is int:
        if len(args) != n_args:
            sys.stderr.write("%s: \"%s\" requires %d arguments but %d "
                             "provided\n"
                             % (argv[0], command_name, n_args, len(args)))
            sys.exit(1)
    else:
        assert False

    func(*args)


if __name__ == '__main__':
    main(sys.argv)



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Defines a Datapath Graph using graphviz. """
import colorsys
import random
import sys

from ovs.flowviz.odp.html import HTMLTree, HTMLFormatter
from ovs.flowviz.odp.tree import FlowTree
from ovs.flowviz.process import FileProcessor

try:
    import graphviz
except ImportError:
    graphviz = None


class GraphProcessor(FileProcessor):
    def __init__(self, opts):
        if graphviz is None:
            print("ERROR: The graph sub-command depends on the graphviz "
                  "Python library, which does not appear to be installed.",
                  file=sys.stderr)
            sys.exit(1)
        super().__init__(opts, "odp")

    def start_file(self, name, filename):
        self.tree = FlowTree()

    def start_thread(self, name):
        pass

    def stop_thread(self, name):
        pass

    def process_flow(self, flow, name):
        self.tree.add(flow, self.opts.get("filter"))

    def process(self):
        super().process(False)

    def print(self, html):
        self.tree.build()

        if len(self.tree.all_recirc_nodes) == 0:
            return

        dpg = DatapathGraph(self.tree, self.opts)
        if not html:
            print(dpg.source())
            return

        html_obj = "<html>"
        html_obj += "<head>"
        html_obj += HTMLTree.head()
        html_obj += "</head>"

        html_obj += "<body>"
        html_obj += HTMLTree.begin_body(self.opts)
        html_obj += "<h1> Flow Graph </h1>"
        html_obj += """<div style="resize:both;
                                   overflow-y:scroll;
                                   overflow-x:scroll;
                                   border: 1px solid;
                                   width:2000px;
                                   height:1200px;">"""
        svg = dpg.pipe(format="svg")
        html_obj += svg.decode("utf-8")
        html_obj += "</div>"
        html_tree = HTMLTree("graph", self.tree, self.opts)
        html_obj += html_tree.format()
        html_obj += HTMLTree.end_body()
        html_obj += "</body>"
        html_obj += "</html>"

        print(html_obj)


class DatapathGraph:
    """A DatapathGraph is a class that renders a set of datapath flows into
    graphviz graphs.

    Args:
        tree: FlowTree
    """

    ct_styles = {}
    node_styles = {
        "default": {
            "style": {},
            "desc": "Default",
        },
        "match": {
            "style": {"color": "#0000ff"},
            "desc": "Flow matches on CT",
        },
        "action": {
            "style": {"color": "#ff0000"},
            "desc": "Flow(s) has CT as action",
        },
    }

    def __init__(self, tree, opts):
        self._tree = tree
        self._opts = opts

        style = HTMLFormatter(self._opts).style

        self.bgcolor = (
            style.get("background").color
            if style.get("background")
            else "#f0f0f0"
        )
        self.fgcolor = (
            style.get("default").color if style.get("default") else "black"
        )

        self._output_nodes = []
        self._graph = graphviz.Digraph(
            "DP flows",
            node_attr={
                "shape": "rectangle",
                "fontcolor": self.fgcolor,
                "color": self.fgcolor,
            },
            edge_attr={
                "color": self.fgcolor,
            },
        )
        self._graph.attr(color=self.fgcolor)
        self._graph.attr(fontcolor=self.fgcolor)
        self._graph.attr(bgcolor=self.bgcolor)
        self._graph.attr(compound="true")
        self._graph.attr(rankdir="LR")
        self._graph.attr(ranksep="3")

        self._populate_graph()

    def source(self):
        """Return the graphviz source representation of the graph."""
        return self._graph.source

    def pipe(self, *args, **kwargs):
        """Output the graph based on arguments given to graphviz.pipe."""
        return self._graph.pipe(*args, **kwargs)

    @classmethod
    def recirc_cluster_name(cls, node):
        """Name of the recirculation cluster."""
        return "cluster_recirc_{}_{}".format(hex(node.recirc), node.in_port)

    @classmethod
    def inport_cluster_name(cls, inport):
        """Name of the input port cluster."""
        return "cluster_inport_{}".format(inport)

    @classmethod
    def invis_node_name(cls, cluster_name):
        """Name of the invisible node."""
        return "invis_{}".format(cluster_name)

    @classmethod
    def output_node_name(cls, port):
        """Name of the ouput node."""
        return "output_{}".format(port)

    @classmethod
    def block_node_name(cls, block):
        """Name of the flow block node."""
        return "flow_block_{}".format(block.flows[0].flow.id)

    def _block_node(self, block):
        """Returns the dictionary of attributes of a graphviz node that
        represents the FlowBlock."""

        url = "#block_{}".format(block.flows[0].flow.id)

        # Graphviz supports HTML-ish syntax. Use it to create a table and
        # place each (summarized) match in each row and the (also summarized)
        # action in the last one.
        # A port is added to the actions row called "actions" that helps make
        # edges start from it.
        label = (
            """<<table border="0" """
            + """cellborder="1" """
            + """cellspacing="0" """
            + """cellpadding="4">"""
        )

        for i, tflow in enumerate(block.flows):
            style = "default"

            # Full representation of the flow to be used in the tooltip.
            full = (
                tflow.flow.section("info").string
                + " "
                + tflow.flow.section("match").string
            )

            # Summarized match: comma separated list of match keys with
            # their ommitted with "..." or "---" depending on whether the
            # match is the same for the entire block or not.
            flowstr = ""
            equal_keys = [m[1].key for m in block.equal_match]
            for m in tflow.flow.match_kv:
                fill = "..."
                if m.key in equal_keys:
                    fill = "---"

                flowstr += (
                    m.meta.kstring + m.meta.delim + fill + m.meta.end_delim
                )

                if m.key == "ct_state" and m.value != "0/0":
                    style = "match"

                flowstr += ","

            flowstr.strip(",")
            color = self.node_styles.get(style)["style"].get(
                "color", self.fgcolor
            )

            label += f"""<tr>
            <td tooltip="{full}" href="{url}" color="{color}">{flowstr}</td>
            </tr>"""

        # Add a row for the action.
        fullact = "actions: " + block.flows[0].flow.section("actions").string
        actstr = ",".join([a.key for a in block.flows[0].flow.actions_kv])

        has_ct_action = bool(
            next(
                filter(
                    lambda x: x.key in ["ct", "ct_clear"],
                    block.flows[0].flow.actions_kv,
                ),
                None,
            )
        )

        style = "action" if has_ct_action else "default"

        color = self.node_styles.get(style)["style"].get("color", self.fgcolor)

        label += f"""<tr>
            <td cellpadding="4"
                border="2"
                href="{url}"
                tooltip="{fullact}"
                color="{color}"
                port="actions">"""
        label += f"<B>actions:</B>  {actstr}"
        label += "</td></tr>"

        label += "</table>>"

        return {
            "name": self.block_node_name(block),
            "label": label,
            "fontsize": "10",
            "nojustify": "true",
            "URL": url,
        }

    def _create_recirc_cluster(self, node):
        """Create a cluster for the RecircNode."""

        cluster_name = self.recirc_cluster_name(node)

        label = "<<B>[recirc 0x{:0x} in_port {}]</B>>".format(
            node.recirc, node.in_port
        )

        cluster = self._graph.subgraph(name=cluster_name, comment=label)
        with cluster as sg:
            sg.attr(rankdir="TB")
            sg.attr(ranksep="0.02")
            sg.attr(label=label)
            sg.attr(margin="5")
            self._add_blocks_to_graph(sg, node.visible_blocks())

        self.processed_recircs.append((node.recirc, node.in_port))

    def _add_blocks_to_graph(self, graph, blocks):
        """Add FlowBlock objects in interable to the graph."""

        # Create an invisible node and an edge to the first block so that
        # it ends up at the top of the cluster.
        invis = self.invis_node_name(graph.name)
        graph.node(invis)
        graph.node(
            invis,
            color=self.bgcolor,
            len="0",
            shape="point",
            width="0",
            height="0",
        )
        first = True

        for block in blocks:
            graph.node(**self._block_node(block))
            if first:
                with graph.subgraph() as c:
                    c.attr(rank="same")
                    c.edge(self.block_node_name(block), invis, style="invis")
                first = False

            # Determine next hop based on block actions.
            self._set_next_node_from_block(block)

    def _set_next_node_from_block(self, block):
        """Create edges to other nodes based on the block's next RecircNodes
        and special actions."""
        created = False

        # Start edges from the "actions" port of the block node.
        name = self.block_node_name(block) + ":actions"

        # Deal with RecircNodes first.
        for node in block.next_recirc_nodes:
            # If the target recirculation cluster has not yet been created,
            # do it now.
            if (node.recirc, node.in_port) not in self.processed_recircs:
                self._create_recirc_cluster(node)

            cname = self.recirc_cluster_name(node)
            self._graph.edge(
                name,
                self.invis_node_name(cname),
                lhead=cname,
                _attributes={"weight": "20"},
            )
            created = True

        # Then, deal with special actions.
        created |= self._set_next_node_from_actions(
            name, block.flows[0].flow.actions
        )

        if not created:
            self._graph.edge(name, "end")

    def _set_next_node_from_actions(self, name, actions):
        """Create edges to other nodes based on the action list."""
        created = False

        for action in actions:
            key, value = next(iter(action.items()))
            if key == "check_pkt_len":
                created |= self._set_next_node_from_actions(
                    name, value.get("gt")
                )
                created |= self._set_next_node_from_actions(
                    name, value.get("le")
                )
            elif key == "sample":
                created |= self._set_next_node_from_actions(
                    name, value.get("actions")
                )
            elif key == "clone":
                created |= self._set_next_node_from_actions(
                    name, value.get("actions")
                )
            else:
                created |= self._set_next_node_action(name, key, value)
        return created

    def _set_next_node_action(self, name, action_name, action_obj):
        """Based on the action object, set the next node."""
        if action_name == "output":
            port = action_obj.get("port")
            if port not in self._output_nodes:
                self._output_nodes.append(port)
            self._graph.edge(
                name, self.output_node_name(port), _attributes={"weight": "1"}
            )
            return True
        elif action_name in ["drop", "userspace", "controller"]:
            if action_name not in self._output_nodes:
                self._output_nodes.append(action_name)
            self._graph.edge(name, action_name, _attributes={"weight": "1"})
            return True
        elif action_name == "ct":
            zone = action_obj.get("zone", 0)
            node_name = "CT zone {}".format(action_obj.get("zone", "default"))
            if zone not in self.ct_styles:
                # Pick a random (highly saturated) color.
                (r, g, b) = colorsys.hsv_to_rgb(random.random(), 1, 1)
                color = "#%02x%02x%02x" % (
                    int(r * 255),
                    int(g * 255),
                    int(b * 255),
                )
                self.ct_styles[zone] = color
                self._graph.node(node_name, color=color)

            color = self.ct_styles[zone]
            self._graph.edge(name, node_name, style="dashed", color=color)
            name = node_name
            return True
        return False

    def _populate_graph(self):
        """Populate the the internal graph."""
        self.processed_recircs = []

        # RecircNode clusters are created recursively when an edge is found
        # pointing to them. Therefore, starting with recirc = 0 nodes.
        for node in self._tree.recirc_nodes.get(0).values():
            if node.visible:
                self._create_recirc_cluster(node)

        # Create an input node that points to each input subgraph
        # They are all inside an anonymous subgraph so that they can be
        # alligned.
        with self._graph.subgraph() as s:
            s.attr(rank="same")
            for inport, node in self._tree.recirc_nodes.get(0).items():
                if not node.visible:
                    continue

                node_name = "input_{}".format(inport)
                cluster_name = self.recirc_cluster_name(node)
                s.node(
                    node_name,
                    shape="Mdiamond",
                    label="input port {}".format(inport),
                )
                self._graph.edge(
                    node_name,
                    self.invis_node_name(cluster_name),
                    lhead=cluster_name,
                    _attributes={"weight": "20"},
                )

        # Create the output nodes in a subgraph so that they are aligned.
        with self._graph.subgraph() as s:
            for port in self._output_nodes:
                s.attr(rank="same")
                if port == "drop":
                    s.node(
                        "drop",
                        shape="Msquare",
                        color="red",
                        label="DROP",
                        rank="sink",
                    )
                elif port == "controller":
                    s.node(
                        "controller",
                        shape="Msquare",
                        color="blue",
                        label="CONTROLLER",
                        rank="sink",
                    )
                elif port == "userspace":
                    s.node(
                        "userspace",
                        shape="Msquare",
                        color="blue",
                        label="CONTROLLER",
                        rank="sink",
                    )
                else:
                    s.node(
                        self.output_node_name(port),
                        shape="Msquare",
                        color="green",
                        label="Port {}".format(port),
                        rank="sink",
                    )

        # Print node style legend.
        with self._graph.subgraph(name="cluster_legend") as s:
            s.attr(label="Legend")
            for style in self.node_styles.values():
                s.node(name=style.get("desc"), _attributes=style.get("style"))
